#!/usr/bin/python
import asyncio
import fp_globals, fp_sqlite
import mpy_wrapper
import new_workers
from fe_types import *


def startWithArgparse():
    import argparse
    fg = fp_globals
    desc = "%s (%s %s)\n%s" % (fg.productName, fg.productChannel, str(fg.productVersion), fg.productCopyright)
    parser = argparse.ArgumentParser(description=desc, formatter_class=argparse.RawTextHelpFormatter,
                                     epilog="THE CLI INTERFACE HAS LIKE 2 COMMANDS THAT GO ANYWHERE. YOU PROBABLY MEANT TO USE startgui!")
    parser.add_argument("--create-db", action="store_true",
                        help="Skip interactive prompt to create database when --dbpath is set from the default.")
    parser.add_argument("--dbpath", type=str,
                        help="Directory to look for/create the database and cache (with --create-db)")

    commonSubcommands = {"list", "show", "statuses", "purge"}

    subparsers = parser.add_subparsers(dest='clicmd')
    commonParser = argparse.ArgumentParser(add_help=False)
    commonParser.add_argument("--json", action="store_true",
                              help="Return output in a \"custom\" JSON format that's likely to change")

    actorParser = subparsers.add_parser('actor', parents=[commonParser])

    accountParser = subparsers.add_parser('account', parents=[commonParser])
    accountParser.add_argument("subcmd", nargs=1, choices=commonSubcommands | {"add", "test"})

    instanceParser = subparsers.add_parser('instance', parents=[commonParser])
    instanceParser.add_argument("subcmd", nargs=1, choices=commonSubcommands | {"directory", "actors", "exclude"})

    guiParser = subparsers.add_parser('startgui')
    return parser.parse_known_args()


async def coreLoop(app):
    import fe_gui
    from fe_core.workers import initWorkers
    asyncio.create_task(checkForUpdates())
    fp_sqlite.initSQLite()
    app.config = {"autoSaveDatabase": autoSaveDatabaseConfigObject(),
                  "saveDatabaseOnExit": saveDatabaseOnExitConfigObject(),
                  "savePersistentTabsOnExit": fe_gui.savePersistentTabsOnExitConfigObject()}
    fp_sqlite.loadAccounts()

    for i in fp_globals.loadedAccounts.keys():
        currentAccount = fp_globals.loadedAccounts[i]
        currentAccountHandle = i
        await currentAccount.log_in()

    mediaPool = new_workers.initWorkers(app)
    initWorkers()

    await fe_gui.guiStartup(app)

    while True:
        slept = 0
        while slept < 60:
            app.updateStatusBar()
            await asyncio.sleep(.5)
            slept = slept + .5
        if app.config["autoSaveDatabase"].configValue:
            fp_sqlite.SAVE()


async def checkForUpdates():
    from httpx import AsyncClient
    try:
        async with AsyncClient() as client:
            response = await client.get("https://fediengine.org/versions.json")
            if response.json()[fp_globals.productChannel]["latest"]["version"] > fp_globals.productVersion:
                print("!!! This version of fediEngine sucks and you should get a new one.")
                if args.gui:
                    fp_globals.Application.showNewVersionAvailableDialog(response.json()[fp_globals.productChannel]["latest"])
    except:
        pass


def launchQt5GUI():
    from sys import argv
    from PyQt5.QtCore import Qt
    from PyQt5.QtWidgets import QApplication
    from qasync import QEventLoop
    from fe_gui import App

    if hasattr(Qt, 'AA_EnableHighDpiScaling'):
        QApplication.setAttribute(Qt.AA_EnableHighDpiScaling, True)
    # if hasattr(Qt, 'AA_UseHighDpiPixmaps'):
    #    QApplication.setAttribute(Qt.AA_UseHighDpiPixmaps, True)

    loop = QEventLoop(QApplication(argv))
    loop.set_debug(True)
    asyncio.set_event_loop(loop)

    fp_globals.Application = App()
    loop.run_until_complete(coreLoop(fp_globals.Application))


async def discoverInstance(domain, quiet=True):
    nativeMetadata, resp = await mpy_wrapper.mpyDiscoverInstance(domain)
    instanceHandle = resp["uri"].lower().replace("https://", "").strip("/")
    try:
        iObject = getInstanceObjectByHandle(instanceHandle) #fucking vanity domains
        if iObject is None:
            raise LookupError
    except LookupError:
        print("#Instance not in RAM or database")
        iObject = InstanceObject(instanceHandle, domain)
        fp_globals.loadedInstances[iObject.handle] = iObject

    iObject._dirty = True
    iObject.meta = nativeMetadata
    im = iObject.meta
    mt = InstanceMetadataType

    try:
        backend = im[mt.ADVERTISED_BACKEND]
    except KeyError:
        backend = "Mastodon"

    if not quiet:
        print("Finished discovering %s at %s" % (iObject.handle, iObject.fulldomain))
        print("\n%s - %s" % (im[mt.TITLE], im[mt.DESCRIPTION]))
        try:
            print("Is run by %s (%s)" % (im[mt.ADMIN_HANDLE], im[mt.ADMIN_EMAIL]))
        except KeyError:
            print("Is run by %s" % im[mt.ADMIN_EMAIL])
        print("Registrations are %s and%s require manual approval by an admin." %
              ({True:"open", False:"closed"}[im[mt.REGISTRATIONS_OPEN]],
               {True:"", False:" do not"}[im[mt.REGISTRATIONS_REQUIRE_APPROVAL]]))
        print("Appears to run %s version %s [%s]" % (backend, im[mt.ADVERTISED_BACKEND_VERSION], im[mt.MASTODON_COMPATIBLE_VERSION]))

    return iObject


class autoSaveDatabaseConfigObject(ConfigObject):
    name = "autoSaveDatabase"
    displayString = "Periodically sync the local database to disk"
    selectionType = "bool"
    default = True


class saveDatabaseOnExitConfigObject(ConfigObject):
    name = "saveDatabaseOnExit"
    displayString = "Sync the database when exiting"
    selectionType = "bool"
    default = False


class defaultAccountConfigObject(ConfigObject):
    name = "defaultAccount"
    displayString = "Account to use by default for shortcuts and account-specific tasks"
    selectionType = "dynamicChoice"
    dynamicChoiceList = fp_globals.loadedAccounts
    default = None

if __name__ == '__main__':
    args, extras = startWithArgparse()
    cmd = args.clicmd
    if cmd in (None, 'startgui'):
        launchQt5GUI()
    else:
        sub = args.subcmd[0]
        fp_sqlite.initSQLite(quiet=True)
        from fe_core.fe_types.actor import ActorObject, getActorObjectByURI, getActorObjectByHandle

        if cmd == 'instance':
            if sub == 'show':
                asyncio.run(discoverInstance(extras[0], quiet=False))
        elif cmd == 'actor':
            print(extras)
            if sub == 'statuses':
                from fe_core.mpy_wrapper import mastodonTimeline
                try:
                    context = getActorObjectByHandle(extras[0])
                except LookupError:
                    context = extras[0]
                result = asyncio.run(mastodonTimeline(context=context))
                print(result)