import warnings
from asyncio.events import get_event_loop
from concurrent.futures.process import ProcessPoolExecutor
from enum import Enum
from time import time
from urllib.parse import urlparse

import attr
import html2text

import fp_sqlite, fp_globals
from .common import MetadataAttribute, thumbnailImage
from .instance import getInstanceObjectByHandle, getInstanceObjectByDomain


@attr.s
class ActorObject(object):
    # Your URI makes you unique
    # And I like you just the way you are.
    # Even you, anon.
    uri = attr.ib()
    userName = attr.ib()
    handle = attr.ib()
    lastSeen = attr.ib(default=0)
    lastUpdated = attr.ib(default=0)
    explicitInstance = attr.ib(default=None)
    _dirty = attr.ib(default=True)

    def __attrs_post_init__(self):
        self.ids = {}
        self.widgets = {}
        self.meta = {}
        get_event_loop().create_task(self.prefetchAndScaleIcon())

    def setMeta(self, target, value, ts=time()):
        try:
            self.meta[target].data = value
        except KeyError:
            self.meta[target] = MetadataAttribute(value, ts, ts)

    async def prefetchAndScaleIcon(self):
        try:
            blob = self.meta[ActorMetadataType.ICON_STATIC_BLOB].data
            return blob
        except KeyError:
            try:
                iconurl = self.meta[ActorMetadataType.ICON_STATIC_URL].data
            except (AttributeError, KeyError):
                return False
            fullblob = fp_sqlite.fetchCachedImageIfExists(iconurl)
            if fullblob is None:
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    #TODO: error handling
                    try:
                        resp = await self.instance.httpxClient.get(iconurl)
                    except Exception as e:
                        print("!!! Skipping actor avatar download for %s due to %s" % (self.handle, repr(e)))
                        return False
                    fullblob = resp.read()

            formatMap = {"jpg": "JPEG",
                         "jpeg": "JPEG",
                         "png": "PNG",
                         "gif": "GIF"}
            imageFormat = formatMap[iconurl.split('?')[0].split('.')[-1].lower()]
            app = fp_globals.Application
            loop = get_event_loop()
            blob = await loop.run_in_executor(app.mediaPool, thumbnailImage, fullblob, imageFormat, 50)
            self.setMeta(ActorMetadataType.ICON_STATIC_BLOB, blob.getvalue())
            loop.run_in_executor(app.ioPool, fp_sqlite.storeImageInCache, iconurl, blob.getvalue())
            return blob

    @property
    def summary(self):
        try:
            return self.meta[ActorMetadataType.SUMMARY].data
        except KeyError:
            return None

    @property
    def displayName(self):
        try:
            return self.meta[ActorMetadataType.DISPLAYNAME].data
        except KeyError:
            return self.userName

    @property
    def sanitizedSummary(self):
        try:
            summary = self.meta[ActorMetadataType.SUMMARY].data
            h = html2text.HTML2Text()
            h.ignore_links = True
            return h.handle(summary).strip()
        except KeyError:
            return '(No summary is available for this Actor)'

    @property
    def domain(self):
        return self.instance.domain

    @property
    def instance(self):
        def lookupInstance():
            print("WARN: Wasting time looking up an instance for %s" % self.uri)
            if self.handle:
                try:
                    return getInstanceObjectByHandle(self.handle.split("@")[1])
                except LookupError:
                    pass
            try:
                return getInstanceObjectByDomain(urlparse(self.uri).netloc)
            except LookupError:
                return getInstanceObjectByHandle(urlparse(self.uri).netloc)

        # TODO: Potentially set explicitInstance to prevent iterating over every instance every time this is called
        if not self.explicitInstance:
            ilres = lookupInstance()
            if ilres is not None:
                self.explicitInstance = ilres
            else:
                print("WARN!!! %s instance lookup is returning None" % self.handle)
                raise LookupError
        return self.explicitInstance



    def __str__(self):
        if self.handle != None:
            return self.handle
        print("WARN!!! __str__ for Actor will be changed soon!")
        return "@%s@%s" % (self.userName, self.domain)

    def __repr__(self):
        return self.uri


def getActorObjectByURI(uri):
    if uri in fp_globals.loadedActors.keys():
        return fp_globals.loadedActors[uri]
    try:
        actor = fp_sqlite.loadActorObjectByURI(uri)
        for i in fp_sqlite.loadActorMetadata(actor.uri):
            print(i[0])
            if i[0] == 2:
                actor.meta[ActorMetadataType.SUMMARY] = MetadataAttribute(i[1], i[2], [3], dirty=False)
        return actor
    except IndexError:
        raise LookupError


def getActorObjectByHandle(handle):
    #TODO: Otherwise validate handle
    handle = handle.lower()
    #TODO: Make global handle->actor dict or have this search by instance
    #Anything to make it... not have to search every fucking loaded actor
    #This function currently only exists for the CLI, where handles will be used way more than URIs
    for actor in fp_globals.loadedActors.values():
        if actor.handle == handle:
            return actor
        elif actor.handle[1:]:
            print("Actor object doesn't have @ before handle")
            return actor
    else:
        try:
            actor = fp_sqlite.loadActorObjectByHandle(handle)
            for i in fp_sqlite.loadActorMetadata(actor.uri):
                print(i[0])
                if i[0] == 2:
                    actor.meta[ActorMetadataType.SUMMARY] = MetadataAttribute(i[1], i[2], [3], dirty=False)
            return actor
        except IndexError:
            raise LookupError


class ActorMetadataType(Enum):
    DISPLAYNAME = 0
    PREFERREDUSERNAME = 0
    SUMMARY = 2
    ICON = 3
    ICON_BLOB = 4
    ICON_URL = 5
    ICON_STATIC = 6
    ICON_STATIC_BLOB = 7
    ICON_STATIC_URL = 8
    HEADER = 9
    HEADER_BLOB = 10
    HEADER_URL = 11
    HEADER_STATIC = 12
    HEADER_STATIC_BLOB = 13
    HEADER_STATIC_URL = 14
    NOBOTS_OLD = 15