import httpx
import attr
from enum import Enum
from time import time


import fp_globals, fp_sqlite
from .common import ConfigObject, MetadataAttribute

class InstanceBackend(Enum):
    PLEROMA = 1
    MISSKEY = 2
    MASTODON = 3
    UNKNOWN = 99

class InstanceMetadataType(Enum):
    TITLE = 1
    DESCRIPTION = 2
    DETECTED_BACKEND = 13
    ADVERTISED_BACKEND = 14
    DETECTED_BACKEND_VERSION = 17
    ADVERTISED_BACKEND_VERSION = 3
    MASTODON_COMPATIBLE_VERSION = 4
    REGISTRATIONS_OPEN = 5
    REGISTRATIONS_REQUIRE_APPROVAL = 6
    ADMIN_HANDLE = 7
    ADMIN_EMAIL = 8
    LANGUAGES = 9
    ADVERTISED_USER_COUNT = 10
    ADVERTISED_STATUS_COUNT = 11
    ADVERTISED_PEER_COUNT = 12
    THUMBNAIL_URI = 15
    ADVERTISED_STATUS_CHARACTER_LIMIT = 16


@attr.s
class InstanceObject(object):
    # An instance is defined by its handle, **NOT** it's uri or domain.
    # This can be very, very confusing. All of the abstraction should be kept here if possible.
    # The reason any of this matters is because you can have @person@domain.com hosted somewhere
    # that is not domain.com. For example, mastodont.domain.com. This user's URI would, thus, be
    # https://mastodont.domain.com/users/person.
    # The "handle" is domain.com. Anywhere else in fediEngine, the instance refers to the handle.
    # The "domain" is mastodont.domain.com. This should only be used by code actually making requests.
    # In theory, the domain for an instance could change.
    #
    # There's more confusing nomenclature. "Accounts" **ALWAYS** refers to local, logged-in user
    # accounts on this client. Accounts as far as the instance is concerned (you know, other people)
    # are called Actors.
    handle = attr.ib()
    domain = attr.ib()
    _httpxClient = attr.ib(default=None,
                           validator=attr.validators.optional(attr.validators.instance_of(httpx.AsyncClient)))
    _monitorhttpxClient = attr.ib(default=None,
                                  validator=attr.validators.optional(attr.validators.instance_of(httpx.AsyncClient)))

    @domain.default
    def _default_domain_to_handle(self):
        return self.handle

    def __attrs_post_init__(self):
        self.lastStreamingError = 0
        self.lastDirectError = 0

        self.initializeConfig()

    def setMeta(self, target, value, ts=time()):
        try:
            self.meta[target].data = value
        except KeyError:
            self.meta[target] = MetadataAttribute(value, ts, ts)

    def initializeConfig(self):
        self.config = {}
        for i in self.configObjects:
            self.config[i.name] = i(context=self.handle)

    # The last time we have any evidence that this instance was on the Internet at all
    # Any status, or 2XX response from it, for example.
    # TODO: Add more sources
    @property
    def lastSeen(self):
        return self.lastSuccessfulDirectContact

    @property
    def httpxClient(self):
        if self._httpxClient:
            return self._httpxClient
        else:
            self._httpxClient = self.mkhttpxClient()
            return self._httpxClient

    @property
    def accounts(self):
        for i in fp_globals.loadedAccounts:
            if i.instance == self:
                yield i

    @property
    def fulldomain(self):
        return "https://%s/" % self.domain

    def streamingFailed(self, err):
        fp_globals.systemMessage("Streaming for instance %s failed due to %s" % (self.handle, repr(err)))
        self.lastDirectError = time()
        self.lastStreamingError = time()
        del(self._monitorhttpxClient)
        self._monitorhttpxClient = None

    def mkmpyClient(self, **kw):
        from mpy_wrapper import makempyClient
        session = self.mkhttpxClient()
        return makempyClient(self, session=session)

    def mkhttpxClient(self, monitor=False):
        try:
            c = self.config
            if monitor:
                # headers = {}
                headers = {'user-agent': str(c["mastodonStreamingMonitorUserAgent"])}
            else:
                headers = {'user-agent': str(c["userAgent"])}
            #TODO: Add SOCKS proxy transport httpx-socks, add relevant config objects
            return httpx.AsyncClient(headers=headers)
        except Exception as e:
            print("!!!datatypes Problem making httpx client")
            print(repr(e))
            raise

    def __str__(self):
        return self.handle

    class disableStreamingConfigObject(ConfigObject):
        name = "disableStreaming"
        displayString = "Do not use the Mastodon streaming API"
        selectionType = "bool"
        default = False

    class includeInFirehoseConfigObject(ConfigObject):
        name = "includeInFirehose"
        displayString = "Monitor this instance when using Firehose"
        selectionType = "bool"
        default = True

    class mastodonStreamingSecondsToBackoffConfigObject(ConfigObject):
        # Applies to all streaming connections, not just monitors
        name = "mastodonStreamingSecondsToBackoff"
        displayString = "Number of seconds to wait before retrying a failed streaming connection"
        selectionType = "int"
        default = 600

    class userAgentConfigObject(ConfigObject):
        name = "userAgent"
        displayString = "Use this User Agent for outbound requests"
        selectionType = "str"
        default = "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:85.0) Gecko/20100101 Firefox/85.0"

    class mastodonStreamingMonitorUserAgentConfigObject(userAgentConfigObject):
        name = "mastodonStreamingMonitorUserAgent"
        displayString = "Use this User Agent when monitoring the Mastodon streaming API"

    class useTorConfigObject(ConfigObject):
        name = "useTor"
        displayString = "Send all requests to this instance through Tor (localhost:9050)"
        selectionType = "bool"
        default = False

    class mastodonStreamingMonitorUseTorConfigObject(useTorConfigObject):
        name = "mastodonStreamingMonitorUseTor"
        displayString = "Use Tor when monitoring the Mastodon streaming API"

    class maximumConcurrentWorkersPerInstanceConfigObject(ConfigObject):
        name = "maximumConcurrentWorkersPerInstance"
        displayString = "Busy at most this many workers for requests to this instance"
        selectionType = "int"
        default = 3

    configObjects = [disableStreamingConfigObject,
                     userAgentConfigObject,
                     maximumConcurrentWorkersPerInstanceConfigObject,
                     includeInFirehoseConfigObject,
                     mastodonStreamingMonitorUserAgentConfigObject,
                     mastodonStreamingSecondsToBackoffConfigObject]

    # Tried a dictionary comprehension. Ran into a serious scope problem.
    configObjectsByName = {}
    for x in configObjects:
        configObjectsByName[x.name] = x


def getInstanceObjectByDomain(domain):
    for instance in fp_globals.loadedInstances.values():
        if instance.domain == domain:
            return instance
    # fp_sqlite.loadInstanceObjectByDomain()
    raise LookupError


def getInstanceObjectByHandle(handle):
    try:
        return fp_globals.loadedInstances[handle]
    except KeyError:
        try:
            return fp_sqlite.loadInstanceObjectByHandle(handle)
        except LookupError:
            # TODO: Attempt to create instance object by querying the instance directly
            raise