import warnings
from asyncio.events import get_event_loop
from concurrent.futures.process import ProcessPoolExecutor
from datetime import datetime
from random import getrandbits
from time import time
from urllib.parse import urlparse

import attr
import html2text

import fp_globals, fp_sqlite
from . import VisibilityScope, getInstanceObjectByDomain, getInstanceObjectByHandle
from .common import thumbnailImage


class StatusObject():
    def __init__(self, uri):
        self.uri = uri
        #self.actor = None
        self.content = ""
        self.subject = ""
        #TODO: move knownBoosts/knownLikes/knownReplies to boosts/likes/replies
        #Old functionality (a count) should be replaced with max(numberReportedByHomeInstance, numberWeHave)
        self.replies, self.boosts, self.likes = 0, 0, 0
        self.knownBoosts, self.knownFavorites, self.knownReplies = [set() for i in range(3)]
        self.ids = {}
        self.widgets = {}
        self.attachments = []
        self.visibilityScope = VisibilityScope['PUBLIC']
        self.inReplyToUri = None
        self._inReplyToStatus = None
        self.explicitInstance = None
        self.contextLastFetched = 0

        # timestamp:actor for boosts
        self.boostedBy = {}

        self._dirty = True

    @property
    def inReplyToStatus(self):
        if self._inReplyToStatus is None:
            self._inReplyToStatus = getStatusObjectByURI(self.inReplyToUri)
        return self._inReplyToStatus

    def registerWidget(self, widget, dataView, clobber=False):
        # TODO: type check dataView and widget
        # TODO: discover dataView from widget, don't need the argument then
        if dataView not in self.widgets.keys():
            self.widgets[dataView] = []

        if clobber:
            self.widgets[dataView] = []

        # the widgetItem isn't hashable. So it's assigned a random fucking number.
        widget.widgetid = getrandbits(128)
        widgetid = widget.widgetid

        self.widgets[dataView].append(widgetid)

    def widgetIsRegistered(self, widget, dataView):
        # TODO: type check dataView and widget
        # TODO: discover dataView from widget, don't need the argument then
        if type(widget) == int:
            return widget in self.widgets[dataView]
        from PyQt5.QtWidgets import QTreeWidgetItem
        if type(widget) == QTreeWidgetItem:
            return widget.widgetid in self.widgets[dataView]
        else:
            fp_globals.systemMessage(
                "Widget passed to widgetIsRegistered isn't int or widget. It's a %s" % type(widget))

    @property
    def homeid(self, fetch=False):
        # TODO: Make fetch happen
        try:
            return self.ids[self.domain]
        except KeyError:
            return None

    @property
    def domain(self):
        #TODO: Figure out which of these is faster
        try:
            return self.instance.domain
        except:
            return urlparse(self.uri).netloc

    @property
    def widgetFormattedTimestamp(self):
        return datetime.utcfromtimestamp(self.timestamp).strftime("%I:%M:%S %p")

    @property
    def instance(self):
        if self.explicitInstance:
            return self.explicitInstance
        if self.actor.instance:
            return self.actor.instance
        try:
            return getInstanceObjectByDomain(urlparse(self.uri).netloc)
        except LookupError:
            return getInstanceObjectByHandle(urlparse(self.uri).netloc)

    @property
    def sanitizedContent(self):
        # HTML-stripped content suitable for the treewidget
        h = html2text.HTML2Text()
        h.ignore_links = True
        return h.handle(self.content).strip().replace('\n', ' ')

    @property
    def images(self):
        # list of attachments that are images
        return [i for i in self.attachments if i.mediaType == "Image"]

    @property
    def hasImageAttachments(self):
        return any([i["type"] == "Image" for i in self.attachments])

    def __str__(self):
        return self.content

    async def fetchContext(self, deep=True):
        if time() - self.contextLastFetched > 120:
            # TODO: instance types for statuses or domains
            from fe_core.mpy_wrapper import mastodonGetContext
            self.contextLastFetched = time()
            try:
                await mastodonGetContext(self, deep=deep)
            except Exception as e:
                fp_globals.systemMessage("Failed to fetch context for %s due to %s" % (self.uri, repr(e)))
        else:
            return

    async def fetchAndCacheImages(self):
        for i in self.attachments:
            if i.mediaType == "Image":
                if not fp_sqlite.fetchCachedImageIfExists(i.uri):
                    if not self.instance.httpxClient:
                        async with aiohttp.ClientSession() as session:
                            async with session.get(i.uri) as resp:
                                try:
                                    imageData = await resp.read()
                                except:
                                    fp_globals.systemMessage("Failed to get image with throwaway httpx")
                    else:
                        try:
                            with warnings.catch_warnings():
                                warnings.simplefilter("ignore")
                                resp = await self.instance.httpxClient.get(i.uri)
                                imageData = resp.read()
                        except Exception as e:
                            async with aiohttp.ClientSession() as session:
                                async with session.get(i.uri) as resp:
                                    imageData = await resp.read()
                    formatMap = {"jpg": "JPEG",
                                 "jpeg": "JPEG",
                                 "png": "PNG",
                                 "gif": "GIF"}
                    imageFormat = formatMap[i.uri.split('?')[0].split('.')[-1].lower()]
                    loop = get_event_loop()
                    app = fp_globals.Application
                    imageBlob = await loop.run_in_executor(app.mediaPool, thumbnailImage, imageData, imageFormat)
                    loop.run_in_executor(app.ioPool, fp_sqlite.storeImageInCache, i.uri, imageBlob.getvalue())


@attr.s(eq=False)
class ReplyObject(object):
    inReplyToUri = attr.ib()
    uri = attr.ib()

    @property
    def inReplyToStatus(self):
        try:
            return getStatusObjectByURI(self.inReplyToUri)
        except Exception as e: #Need name of it to make it more specific
            print("replyObject exception to grab is %s" % repr(e))
            #TODO: fetch status or raise
            domain = urlparse(self.inReplyToUri).netloc
            try:
                instance = getInstanceObjectByDomain(domain)
                instance.fetchStatus(self.inReplyToUri)
            except LookupError:
                #TODO: use throwaway
                #TODO: return ghost object of status
                return None


def getStatusObjectByURI(uri, instance=None):
    # TODO: Perform instance assignment in fp_sqlite
    if uri in fp_globals.loadedStatuses.keys():
        if instance is not None:
            fp_globals.loadedStatuses[uri].instance = instance
        return fp_globals.loadedStatuses[uri]
    try:
        requestedStatus = fp_sqlite.loadStatusObjectByURI(uri)
        for i in requestedStatus.knownReplies:
            try:
                h = getStatusObjectByURI(i.inReplyToUri)
            except LookupError:
                pass  # Reply isn't in DB. Don't bother
                # TODO: auto-crawl it.
            if requestedStatus.hasImageAttachments:
                # TODO: test for trying to view images before they've been fetched
                get_event_loop().create_task(requestedStatus.fetchAndCacheImages())
        if instance is not None:
            fp_globals.loadedStatuses[uri].instance = instance
        return requestedStatus
    except IndexError:
        raise LookupError


@attr.s
class AttachmentObject(object):
    uri = attr.ib()
    thumburi = attr.ib(default=None)
    description = attr.ib(default=None)
    mediaType = attr.ib(default=None)
    sensitive = attr.ib(default=False)
    _dirty = attr.ib(default=True)

    @property
    def size(self):
        return None

    @property
    def sha512hash(self):
        return None

    @property
    def thumbsha512hash(self):
        return None