#!/usr/bin/python
import sqlite3
from os import path, makedirs
from time import time
from urllib.parse import urlparse
import fp_globals
from fe_types import *
from fe_types.actor import ActorMetadataType
from external.appdirs import user_data_dir


datadir = user_data_dir("fediengine_gui")
mainDBPath = path.join(datadir, "fp_main.db")
mainTemplatePath = path.join("assets", "main_template.db")
imageCachePath = path.join(datadir, "fp_imagecache.db")


def loadStatusObjectByURI(uri):
    with sqlite3.connect(mainDBPath) as sqlc:
        row = sqlc.execute("SELECT * FROM statuses WHERE uri = ?", (uri,)).fetchone()
        if row is None:
            raise LookupError
        else:
            sObject = StatusObject(uri)
            sObject.actor = loadActorObjectByHandle(row[1])
            sObject.timestamp = row[2]
            sObject.lastActivity = row[3]
            sObject.lastChecked = row[4]
            sObject.explicitInstance = loadInstanceObjectByHandle(row[9])
            #TODO: hashtags
            sObject.visibilityScope = VisibilityScope(row[14])
            sObject.content = row[11]
            sObject.subject = row[13]

            loadAttachments = row[5]
            loadReplies = row[6]
            loadBoosts = row[7]
            loadFavorites = row[8]
            #TODO: act upon these
            if loadReplies:
                for i in sqlc.execute("SELECT status_uri FROM status_replies WHERE inreplyto_uri = ?", (uri,)):
                    sObject.knownReplies.add(ReplyObject(uri, i[0]))
                    # TODO: Warn if none were found, they may need to be fetched

            #Just came from DB
            sObject._dirty = False
            fp_globals.loadedStatuses[uri] = sObject
            return fp_globals.loadedStatuses[uri]
    raise IndexError

def globalContentSearch(search):
    sqlc = sqlite3.connect(mainDBPath)
    #TODO: also return flatcontent, and consider indexing it or adding an option to do so
    return sqlc.execute("SELECT uri FROM statuses WHERE flatcontent LIKE '%%?%%'", (search))


def loadInstanceObjectByHandle(targetHandle):
    sqlc = sqlite3.connect(mainDBPath)
    #TODO: Don't use for loop because there will only be a single result
    for row in sqlc.execute("SELECT domain, excluded FROM instances WHERE handle = ?", (targetHandle,)):
        if row is None:
            raise LookupError
        newInstanceObject = InstanceObject(targetHandle, domain=row[0])

        if newInstanceObject.handle in fp_globals.loadedInstances.keys():
            print("!!!SQL Asked to load a handle %s whose instance is already loaded" % (targetHandle))

        fp_globals.loadedInstances[targetHandle] = newInstanceObject
        return newInstanceObject

def loadEveryInstance():
    print("loading every instance")
    sqlc = sqlite3.connect(mainDBPath)
    for row in sqlc.execute("SELECT * FROM instances"):
        print(row)



def loadActorObjectByURI(uri):
    return loadActorObjectBy("SELECT uri, handle, lastseen, lastupdated FROM actors WHERE uri = ?", uri)


def loadActorObjectByHandle(handle):
    return loadActorObjectBy("SELECT uri, handle, lastseen, lastupdated FROM actors WHERE handle = ?", handle)


def loadActorObjectBy(query, target):
    if target.find('@') != -1:
        instance = getInstanceObjectByHandle(target.split('@')[-1])
    else:
        instance = getInstanceObjectByDomain(urlparse(target).netloc)

    with sqlite3.connect(mainDBPath) as sqlc:
        row = sqlc.execute(query, (target,)).fetchone()
        if row is None:
            raise IndexError
        print("sql")
        print(row)
        userName = row[0].split('@')[-1]
        aObject = ActorObject(uri=row[0],
                              explicitInstance=instance,
                              userName=userName,
                              handle=row[1],
                              lastSeen=row[2],
                              lastUpdated=row[3])

        for row in sqlc.execute("SELECT instance, id FROM instances_ids WHERE entity = ?", (aObject.handle,)):
            aObject.ids[row[0]] = row[1]
        # TODO: Fetch or use null placeholder when no result is returned

    aObject._dirty = False
    fp_globals.loadedActors[row[0]] = aObject


def loadActorMetadata(actoruri):
    #TODO: Allow narrowing by key or list of keys
    #TODO: Return ghost objects
    with sqlite3.connect(mainDBPath) as sqlc:
        return [(ActorMetadataType(row[0]), row[1]) for row in sqlc.execute("SELECT metatype, data, firstseen, lastseen FROM actors_metadata WHERE actor = ?", (actoruri,))]


def syncInstances(instances: iter, sqlc: sqlite3.Connection = None):
    if sqlc is None:
        sqlc = sqlite3.connect(mainDBPath)

    iQueries = set()
    imQueries = set()
    flushedInstances = []
    for instance in instances:
        iQueries.add((instance.handle,
                        instance.domain,
                        False,
                        0,
                        0,
                        0))
    try:
        with sqlc:
            sqlc.executemany("INSERT INTO instances VALUES (?, ?, ?, ?, ?) ON CONFLICT(handle) DO UPDATE SET lastseen=?", iQueries)
    except (sqlite3.OperationalError, sqlite3.IntegrityError) as e:
        fp_globals.systemMessage("sqlite3 instance flush error: %s" % repr(e))
        raise


def syncActors(actors: iter, sqlc: sqlite3.Connection = None):
    if sqlc is None:
        sqlc = sqlite3.connect(mainDBPath)

    aQueries = set()
    amQueries = set()
    flushedActors = []
    for actor in actors:
        try:
            #TODO: Update bios and other information in actors_metadata
            aQueries.add((actor.handle,
                          actor.uri,
                          actor.instance.handle,
                          actor.displayName,
                          actor.lastSeen,
                          int(time()),
                          int(time())))

            amQueries.update([(actor.handle, x.value, y.fs, y.ls, None, y.v) for x, y in actor.meta.items() if y.dirty])

            #Wait to mark object clean until database commit succeeds
            flushedActors.append(actor)
        except Exception as e:
            print("skipping %s due to %s" % (actor.handle, repr(e)))
    try:
        with sqlc:
            sqlc.executemany("INSERT INTO actors VALUES (?, ?, ?, ?, ?, ?) ON CONFLICT(handle) DO UPDATE SET lastupdated=?", aQueries)
            sqlc.executemany("REPLACE INTO actors_metadata VALUES (?, ?, ?, ?, ?, ?)", amQueries)
    except (sqlite3.OperationalError, sqlite3.IntegrityError) as e:
        print(repr(e))
        fp_globals.systemMessage("sqlite3 actor flush error: %s" % repr(e))
        raise #TODO: continue but don't write statuses etc for unflushed actors
    for actor in flushedActors:
        actor._dirty = False


def syncStatuses(statuses: iter, sqlc: sqlite3.Connection = None):
    if sqlc is None:
        sqlc = sqlite3.connect(mainDBPath)

    statusQueries, boostsQueries, favoritesQueries, repliesQueries, attachmentsQueries = [[] for i in range(5)]
    flushedStatuses, flushedBoosts, flushedFavorites, flushedReplies, flushedAttachments = [[] for i in range(5)]
    for status in statuses:
        try:
            #TODO: serialize content
            #TODO: boosts/favorites as list of objects@statusObject
            #TODO: Hashtags, privacy scope. placeholders are None 09jan2020
            #TODO: Use real lastActivity instead of status timestamp
            statusQueries.append((status.uri,
                                  status.actor.handle,
                                  int(status.timestamp),
                                  int(status.timestamp),
                                  int(status.contextLastFetched),
                                  0 < len(status.attachments),
                                  0 < len(status.knownReplies),
                                  0 < len(status.knownBoosts),
                                  0 < len(status.knownFavorites),
                                  status.instance.handle,
                                  None, status.visibilityScope.value,
                                  status.content,
                                  status.sanitizedContent,
                                  status.subject,
                                  int(status.contextLastFetched)))

            if len(status.knownBoosts) > 0:
                for boost in status.knownBoosts:
                    if boost._dirty:
                        #TODO: verify or create uniqueness constraint
                        boostsQueries.append((status.uri,
                                             boost.actor,
                                             boost.timestamp))
                        flushedBoosts.append(boost)

            if len(status.knownFavorites) > 0:
                for favorite in status.knownFavorites:
                    if favorite._dirty:
                        #TODO: verify or create uniqueness constraint
                        favoritesQueries.append((status.uri,
                                                favorite.actor,
                                                favorite.timestamp))
                        flushedFavorites.append(favorite)

            if len(status.attachments) > 0:
                for attachment in status.attachments:
                    if attachment._dirty:
                        try:
                            mediaType = {'Image': 1, None: None}[attachment.mediaType]
                        except KeyError:
                            mediaType = None

                        attachmentsQueries.append((status.uri,
                                                   attachment.uri,
                                                   attachment.thumburi,
                                                   mediaType,
                                                   attachment.sensitive,
                                                   attachment.description,
                                                   attachment.size,
                                                   attachment.sha512hash,
                                                   attachment.thumbsha512hash))
                        flushedAttachments.append(attachment)

            if len(status.knownReplies) > 0:
                for reply in status.knownReplies:
                    if reply.uri in fp_globals.loadedStatuses.keys():
                        #Try to find it in DB? Do replies separately?
                        pass

            flushedStatuses.append(status)
        except Exception as e:
            print("skipping %s due to %s" % (repr(status), repr(e)))

    try:
        with sqlc:
            #TODO: Update more than one column or write new queries separately
            sqlc.executemany("INSERT INTO statuses (uri, actor, timestamp, lastactivity, lastchecked, hasattachments, hasreplies, hasboosts, hasfavorites, instance, hashtags, visibilityscope, content, flatcontent, subject) VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?) ON CONFLICT (uri) DO UPDATE SET lastchecked=?", statusQueries)
    except (sqlite3.OperationalError, sqlite3.IntegrityError) as e:
        fp_globals.systemMessage("sqlite3 status/status flush error: %s" % repr(e))
        raise

    try:
        with sqlc:
            sqlc.executemany("INSERT INTO statuses_attachments VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?) ON CONFLICT (uri) DO NOTHING", attachmentsQueries)
        for attachment in flushedAttachments:
            attachment._dirty = False
    except (sqlite3.OperationalError, sqlite3.IntegrityError) as e:
        fp_globals.systemMessage("sqlite3 status/attachments flush error: %s" % repr(e))
        raise

    #Statuses only marked clean if all other object inserts finished without errors
    for status in flushedStatuses:
        status._dirty = False


def syncDirtyObjects():
    #TODO: optional list of which objects to flush
    #TODO: clock individual operations
    sqlc = sqlite3.connect(mainDBPath)

    syncInstances((x for x in fp_globals.loadedInstances.values()), sqlc)
    syncActors((x for x in fp_globals.loadedActors.values() if x._dirty), sqlc)
    syncStatuses((x for x in fp_globals.loadedStatuses.values() if x._dirty), sqlc)

    sqlc.close()


def syncConfigObjects(configObjects=None, sqlc=None, c=None, force=False):
    # TODO: see if there's a built in iterable of all (sub)class instances
    if configObjects is None:
        configObjects = []
        for instance in fp_globals.loadedInstances.values():
            configObjects.extend(instance.config.values())
    elif type(configObjects) != list:
        if isinstance(configObjects, ConfigObject):
            configObjects = [configObjects]
        else:
            print("syncConfigObjects was passed a %s" % type(configObjects))
            raise TypeError

    if sqlc is None:
        sqlc = sqlite3.connect(mainDBPath)
        c = sqlc.cursor()

    written = []
    # TODO: refactor to use executemany()
    for co in configObjects:
        if force or co.dirty:
            try:
                cv = co.configValue
            except AttributeError:  # Class/default
                if co.context == "_global_":  # santy check
                    if co in fp_globals.dirtyGlobalConfigObjects:
                        cv = co.default
                else:
                    print("!!!SQL WARN content object %s has no configValue but also isn't a default/class!" % repr(co))
                    continue
                c.execute('DELETE FROM config WHERE context=? AND key=?', (co.context, co.name))
                c.execute('INSERT INTO config VALUES (?,?,?)', (co.context, co.name, cv))
                written.append(co)

    sqlc.commit()
    for co in written:
        if co.context == "_global_":
            fp_globals.dirtyGlobalConfigObjects.remove(co)
        else:
            co.dirty = False
    sqlc.close()


def loadConfigValue(targetConfigName, context="_global_"):
    sqlc = sqlite3.connect(mainDBPath)
    cursor = sqlc.cursor()
    res = cursor.execute("SELECT value FROM config WHERE key = ? AND context = ?",
                         (targetConfigName, context)).fetchall()
    if len(res) > 1:
        print("WARNING: More than one result for config search %s with context %s" % (
            targetConfigName, context))
    elif len(res) == 0:
        raise LookupError
    else:
        print("+++SQL got response for %s" % targetConfigName)
        return res[0]


def savePersistentTabs():
    # I looked for a very long time for an iterable list of tabs
    tabWidget = fp_globals.Application.tabs
    tabsToSave = []
    for i in range(tabWidget.count()):
        if tabWidget.widget(i).systemTab:
            continue
        if tabWidget.widget(i).config["tabIsPersistent"].configValue:
            tabsToSave.append(tabWidget.widget(i))

    print("There are %i tabs to save" % len(tabsToSave))

    queries = []
    for position, tab in enumerate(tabsToSave, start=0):
        print(type(tab).__name__)

        queries.append((str(tab.configContext),
                        str(type(tab).__name__),
                        str(tab.context)))

    print(queries)

    sqlc = sqlite3.connect(mainDBPath)
    c = sqlc.cursor()
    c.execute("DELETE FROM persistent_tabs")
    c.execute("DELETE FROM config WHERE context LIKE '_tab_%'")
    c.executemany("INSERT INTO persistent_tabs (tab, tabclass, context) Values (?,?,?)", queries)

    # Commit and close is handled by syncConfigObjects
    syncConfigObjects(sum([list(tab.config.values()) for tab in tabsToSave], []), sqlc, c, True)


def loadPersistentTabs():
    with sqlite3.connect(mainDBPath) as sqlc:
        return sqlc.execute("SELECT * FROM persistent_tabs").fetchall()


def loadAccounts():
    with sqlite3.connect(mainDBPath) as sqlc:
        for i in sqlc.execute("SELECT * FROM accounts").fetchall():
            fp_globals.loadedAccounts[i[0]] = AccountObject(handle=i[0],
                                                            auth=tuple(i[1:5]),
                                                            enabled=i[5],
                                                            api_path=i[6])


def addAccount(handle, client_id, client_secret, oauth_code, access_token):
    sqlc = sqlite3.connect(mainDBPath)
    c = sqlc.cursor()
    c.execute("INSERT INTO accounts (handle, client_id, client_secret, oauth_code, access_token) Values (?,?,?,?,?)",
              (handle, client_id, client_secret, oauth_code, access_token))
    sqlc.commit()
    sqlc.close()


def fetchCachedImageIfExists(imageUrl):
    sqlc = sqlite3.connect(imageCachePath)
    c = sqlc.cursor()
    h = c.execute("SELECT data FROM 'imagecache' WHERE url='%s'" % imageUrl).fetchall()
    if len(h) == 0:
        return None
    else:
        return h[0][0]


def storeImageInCache(url, imageBlob):
    sqlc = sqlite3.connect(imageCachePath)
    c = sqlc.cursor()
    c.execute("INSERT INTO imagecache (url, data, added) Values (?,?,?)", (url, imageBlob, time()))
    sqlc.commit()
    sqlc.close()


def SAVE():
    started = time()
    syncDirtyObjects()
    syncConfigObjects()
    print("File saved. Took %s seconds." % "{:.4f}".format(time() - started))


def initImageCache(db_path):
    with sqlite3.connect(db_path) as sqlc:
        sqlc.execute("CREATE TABLE imagecache (url text, data blob, added int)")


def initSQLite(quiet=False):
    # Create main DB if needed
    if path.isfile(mainDBPath):
        if not quiet:
            print("Found existing main DB (%s KB)" % str(path.getsize(mainDBPath)/1024))
        with sqlite3.connect(mainDBPath) as sqlc:
            #This is included in the template, and isn't set by the software.
            dbver = int(sqlc.execute("PRAGMA user_version").fetchone()[0])
        if dbver < fp_globals.databaseVersion:
            from shutil import copyfile, move
            print("The existing database is outdated. It is being moved, and a clean DB will be created in its place.")
            print("As a reminder, the database schema is expected to change with almost every pre-alpha release.")
            #TODO: Migrate accounts table if it has not changed 
            move(mainDBPath, "%s.OLD%i" % (mainDBPath, time()))
            copyfile(mainTemplatePath, mainDBPath)

    else:
        from shutil import copyfile
        # create directory if it doesn't exist
        if not path.isdir(path.dirname(mainDBPath)):
            makedirs(path.dirname(mainDBPath))
        #Initialize main DB by copying template from /assets
        copyfile(mainTemplatePath, mainDBPath)

    # Create image cache DB if needed
    #TODO: Purge old files if cache is too large
    if not path.isfile(imageCachePath):
        # create directory if it doesn't exist
        if not path.isdir(path.dirname(imageCachePath)):
            makedirs(path.dirname(imageCachePath))
        initImageCache(imageCachePath)
    else:
        if not quiet:
            print("Found existing image cache. (%s KB)" % str(path.getsize(imageCachePath)/1024))
