#!/usr/bin/python
import httpx
import warnings
import asyncio
from urllib.parse import urlparse
from time import time
import workers, fp_globals
from fe_core import discoverInstance
from fe_types import *
from fe_types.actor import ActorMetadataType
from fe_types.instance import InstanceMetadataType
from external.streaming import *
from external.mastodon import Mastodon
from external.mastodon import MastodonMalformedEventError, MastodonInternalServerError, MastodonNetworkError, MastodonReadTimeout


class mpyStreamCallback(StreamListener):
    def __init__(self, domain, statusReceiver, heartbeatReceiver=None):
        StreamListener.__init__(self)
        self.heartbeatReceiver = heartbeatReceiver
        self.domain = domain
        self.statusReceiver = statusReceiver
        self.lastActive = -1
        fp_globals.activeStreamingObjects.add(self)

    def onActivity(self):
        self.lastActive = time()
        if self.heartbeatReceiver is not None:
            self.heartbeatReceiver(self.lastActive)

    #Overriden functions from mastodon.py
    def on_abort(self, err):
        #TODO: pop error banner if this connection was started by a tab
        fp_globals.activeStreamingObjects.remove(self)
        self.domain.streamingFailed(err)

    def on_update(self, mpyStatus):
        self.onActivity()
        self.statusReceiver(mpyTootDictToStatusObject(mpyStatus, self.domain))

    def handle_heartbeat(self):
        self.onActivity()


class mpyMonitorStreamCallback(mpyStreamCallback):
    def on_update(self, mpyStatus):
        #Determine if we give a shit
        #Currently checking only by instance, and only by origination

        if fp_globals.Application.firehoseAction.isChecked():
            #do minimum amount of processing necessary to show status
            #do not crawl it or download/thumbnail images

            sObject = mpyTootDictToStatusObject(mpyStatus, self.domain, cacheImageThumbnails=False)
            fp_globals.Application.firehoseTab.addStatus(sObject, fetchContext=False)


class mpyUserStreamCallback(mpyStreamCallback):
    def __init__(self, account, statusReceiver):
        mpyStreamCallback.__init__(self, account.instance.domain, statusReceiver)
        self.account = account
        self.domain = account.instance.domain
        self.statusReceiver = statusReceiver

    def on_notification(self, mpyNotification):
        nativeNotificiation = makempyNotificationNative(mpyNotification, self.domain)
        for i in self.account.notificationTabs:
            i.addNotification(nativeNotificiation, self.account)
        fp_globals.unifiedNotificationsTab.addNotification(nativeNotificiation, self.account)


async def mpyDiscoverInstance(domain):
    #This works with all instances*, even Pleroma, because mastodon.py just converts the JSON
    #response from the server into a python dict. This means any additional fields come for free.
    #But it also means that fields in the documentation aren't guaranteed to be in the dict.
    #"contact_account" and "short_description" seem to be ones that are routinely missing.
    try:
        mpy = await Mastodon(api_base_url=domain)
        resp = await mpy.instance()
    except MastodonNetworkError:
        #TODO: mark instance object with error
        print("Captured remote instance error during discovery")

    #TODO: Webfinger for true api_base_url if version check fails, maybe upstream that

    i = InstanceMetadataType
    nativeMap = ((i.TITLE, "title"),
                 (i.DESCRIPTION, "description"),
                 (i.ADMIN_EMAIL, "email"),
                 (i.REGISTRATIONS_OPEN, "registrations"),
                 (i.THUMBNAIL_URI, "thumbnail"),
                 (i.REGISTRATIONS_REQUIRE_APPROVAL, "approval_required"))

    nativeMetadata = {x[0]:resp.get(x[1], None) for x in nativeMap}

    #TODO: Turn this into a fast dict comprehension
    if 1:
        mstats = resp["stats"]
        nativeMetadata[i.ADVERTISED_STATUS_COUNT] = mstats["status_count"]
        nativeMetadata[i.ADVERTISED_USER_COUNT] = mstats["user_count"]
        nativeMetadata[i.ADVERTISED_PEER_COUNT] = mstats["domain_count"]

    versionString = resp["version"]
    if "pleroma" in resp.keys():
        splitVersion = versionString[:-1].split(" (compatible; Pleroma ")
        nativeMetadata[i.MASTODON_COMPATIBLE_VERSION], nativeMetadata[i.ADVERTISED_BACKEND_VERSION] = splitVersion
        nativeMetadata[i.ADVERTISED_BACKEND] = InstanceBackend.PLEROMA
    else:
        if "compatible" in versionString.lower():
            splitVersion = versionString.split(" ")
            nativeMetadata[i.MASTODON_COMPATIBLE_VERSION] = splitVersion[0]
            nativeMetadata[i.ADVERTISED_BACKEND_VERSION] = splitVersion[1]
        else:
            nativeMetadata[i.MASTODON_COMPATIBLE_VERSION] = resp["version"]
            nativeMetadata[i.ADVERTISED_BACKEND_VERSION] = resp["version"]

    return nativeMetadata, resp


def makempyClient(context, **kw):
        if type(context) == InstanceObject:
            return Mastodon(api_base_url=context.domain, **kw)


def makempyNotificationNative(mpyNotification, domain):
    return {'id': mpyNotification['id'],
            'type': mpyNotification['type'],
            'status': mpyTootDictToStatusObject(mpyNotification['status'], domain),
            'source': mpyUserDictToActorObject(mpyNotification['account'], domain)}


def mpyUserDictToActorObject(mud, domain, ts=time()):
    try:
        actor = getActorObjectByURI(mud['url'])
    except LookupError:
        #Not in memory or DB.
        if '@' not in mud['acct']:
            userName = mud['acct'].lower()
            instance = getInstanceObjectByDomain(domain)
            handle = "%s@%s" % (userName, instance.handle)
        else:
            handle = mud['acct'].lower()
            userName, instanceHandle = handle.split('@')
            instance = getInstanceObjectByHandle(instanceHandle)

        actor = ActorObject(uri=mud['url'],
                            userName=userName,
                            handle=handle,
                            lastSeen=ts,
                            explicitInstance=instance)

    if actor.lastSeen < ts:
        actor.lastSeen = ts
        actor._dirty = True

    fields = (('note', ActorMetadataType.SUMMARY),
              ('display_name', ActorMetadataType.DISPLAYNAME),
              ('avatar', ActorMetadataType.ICON_URL),
              ('avatar_static', ActorMetadataType.ICON_STATIC_URL),
              ('header', ActorMetadataType.HEADER_URL),
              ('header_static', ActorMetadataType.HEADER_STATIC_URL))

    for mpyField, nativeMetadataType in fields:
        #TODO: Verify how mastodon.py denotes an empty value
        if mud[mpyField] is not None:
            try:
                if mud[mpyField] != actor.meta[nativeMetadataType].data:
                    actor.setMeta(nativeMetadataType, mud[mpyField], ts)
            #There's a performance reason to do this. Srsly.
            except KeyError:
                actor.setMeta(nativeMetadataType, mud[mpyField], ts)
    #TODO: Instance object as ids key
    actor.ids[domain] = str(mud['id'])
    try:
        return fp_globals.loadedActors[actor.uri]
    except KeyError:
        fp_globals.loadedActors[actor.uri] = actor
        return fp_globals.loadedActors[actor.uri]


def statusFromMastodonStatusId(targetID, domain):
    targetID = str(targetID)
    # Attempt to find status in memory
    for i in fp_globals.loadedStatuses.values():
        try:
            if i.ids[domain] == targetID:
                return i
        except KeyError:
            continue

    # Attempt to find status in database

    # Attempt to ask this mastodon server for it

    # Attempt to ask the remote mastodon server for it (if applicable)

    # Attempt to ask the remote mastodon server for it via ActivityPub


def mpyTootDictToStatusObject(mtd, mpyDomain, dontParseReplies=False, dontSave=False, cacheImageThumbnails=True):
    # Instance assignment
    if type(mpyDomain) != str:
        try:
            mpyDomain = mpyDomain.domain
        except:
            print("!!!mpy passed a domain that wasn't str")
            raise

    if mtd["visibility"] == "direct":
        return None

    if not mtd['reblog']:
        targetDomain = urlparse(mtd["uri"]).netloc
        try:
            targetInstance = getInstanceObjectByDomain(targetDomain)
        except LookupError:
            #kickoff instance discovery, then make a temporary placeholder object
            discoveryTask = asyncio.create_task(discoverInstance(targetDomain))
            dhandle = mtd['account']['acct'].split('@')[1]
            ddomain = urlparse(mtd['account']['url']).netloc
            targetInstance = InstanceObject(handle=dhandle, domain=ddomain)
            fp_globals.loadedInstances[dhandle] = targetInstance

        targetInstance.lastSuccessfulDirectContact = time()
        try:
            foundStatus = False
            existingStatus = getStatusObjectByURI(mtd['uri'])
            foundStatus = True
            # TODO: Update potentially outdated information

            try:
                testLookup = str(existingStatus.ids[mpyDomain])
                if str(existingStatus.ids[mpyDomain]) != str(mtd["id"]):
                    print("!!!!! About to replace id %s with %s" % (testLookup, mtd["id"]))
            except KeyError:
                pass

            existingStatus.ids[mpyDomain] = str(mtd["id"])

            if mtd['in_reply_to_id']:
                if not dontParseReplies:
                    statusFromId = statusFromMastodonStatusId(mtd["in_reply_to_id"], mpyDomain)
                    if statusFromId:
                        existingStatus.inReplyTo = statusFromId
                        existingStatus.inReplyToUri = statusFromId.uri
                        statusFromId.knownReplies.add(ReplyObject(statusFromId.uri, existingStatus.uri))
            if dontSave:
                existingStatus._dirty = False
            else:
                existingStatus._dirty = True  # assuming
            return existingStatus

        except LookupError:
            if not foundStatus:
                pass  # Create object
            else:
                print("!!!!! Had a lookup error *after* finding a status!")

        sObject = StatusObject(mtd['uri'])

        try:
            sObject.actor = fp_globals.loadedActors[mtd['account']['url']]
        except KeyError:  # actor needs made
            sObject.actor = mpyUserDictToActorObject(mtd['account'], mpyDomain)

        # copy variables
        #TODO: Verify timestamp timezone conversion is correct/actually needed
        sObject.content = mtd['content']
        sObject.subject = mtd['spoiler_text']
        sObject.visibilityScope = VisibilityScope[mtd['visibility'].upper()]
        sObject.timestamp = mtd['created_at'].timestamp()
        sObject.boosts = mtd['reblogs_count']
        sObject.likes = mtd['favourites_count']
        sObject.replies = mtd['replies_count']
        sObject.ids[mpyDomain] = str(mtd['id'])
        sObject.mpyDict = mtd

        for i in mtd["media_attachments"]:
            if i["type"] == "image":
                sObject.attachments.append(AttachmentObject(i.url,
                                                            mediaType="Image",
                                                            description=i.description,
                                                            sensitive=mtd["sensitive"]))

        if cacheImageThumbnails:
            asyncio.create_task(sObject.fetchAndCacheImages())

        if mtd['in_reply_to_id']:
            statusFromId = statusFromMastodonStatusId(mtd["in_reply_to_id"], mpyDomain)
            if statusFromId:
                sObject.inReplyTo = statusFromId
                sObject.inReplyToUri = statusFromId.uri
                statusFromId.knownReplies.add(ReplyObject(statusFromId.uri, sObject.uri))

        # make sure to return a reference to the object in fp_globals.loadedStatuses
        fp_globals.loadedStatuses[mtd['uri']] = sObject
        return fp_globals.loadedStatuses[mtd['uri']]

    else:
        boostedStatus = mpyTootDictToStatusObject(mtd["reblog"], mpyDomain)
        try:
            booster = fp_globals.loadedActors[mtd['account']['url']]
        except KeyError:  # actor needs made
            booster = mpyUserDictToActorObject(mtd['account'], mpyDomain)
        boostedStatus.boostedBy[mtd['created_at']] = booster
        return boostedStatus


async def fetchStatusIdFromInstance(status, instance, mpyClient=None, session=None):
    from fe_core.external.mastodon import MastodonInternalServerError
    domain = instance.domain
    if mpyClient is None:
        try:
            mpyClient = await Mastodon(api_base_url=domain, session=session)
        except Exception as e:
            print("%s while creating mpyclient to fetch status context for domain %s" % (repr(e), domain))
            raise

    try:  # mastodon
        statusSearchResults = [await mpyClient.status(status.uri.split("/")[-1])]
    except MastodonInternalServerError:  # pleroma
        searchResults = await mpyClient.search(status.uri)
        statusSearchResults = searchResults["statuses"]
    for i in statusSearchResults:  # mpy status dicts
        if i.uri == status.uri:
            sObject = mpyTootDictToStatusObject(i, domain, dontParseReplies=True)
            try:
                status_id = sObject.ids[instance.domain]
            except KeyError:
                print("Status ID lookup failed after refetching status from instance directly!")
                raise

    return status_id



async def mastodonGetContext(status, domain=None, deep=True):
    def mpyCallback(*args, **kwargs):
        pass

    try:
        instance = status.instance
        domain = instance.domain
        if not instance.httpxClient:
            instance.httpxClient = httpx.AsyncClient()
        session = instance.httpxClient
    except LookupError:
        print("%s has no instance object yet, bailing." % domain)
        #session = httpx.AsyncClient()
        raise

    try:
        mpyClient = await Mastodon(api_base_url=domain, session=session)
    except Exception as e:
        print("%s while creating mpyclient to fetch status context for domain %s" % (repr(e), domain))

    try:
        try:
            status_id = status.ids[domain]
        except KeyError:
            status_id = await fetchStatusIdFromInstance(status, instance)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            mpyContextDict = await mpyClient.status_context(status_id)
    except ConnectionRefusedError as e:
        print("Connection refused for instance %s at domain %s" % (instance, instance.domain))
        return
    except Exception as e:
        fp_globals.systemMessage("Skipping %s due to %s" % (instance.domain, repr(e)), sender="MastodonContextFetcher",
                                 severity="WARN")
        return

    for reply in mpyContextDict["descendants"]:
        try:
            h = mpyTootDictToStatusObject(reply, instance.domain)
            if deep and (time() - h.contextLastFetched) < 120:
                workers.schedulempyJob(h.fetchContext, {}, "statusCrawlQueue", fp_globals.doFuckAll, target="anonymous")
        except Exception as e:
            print("mpyerror: %s" % repr(e))


async def startStreamingThread(context, streamType, statusReceiver, heartbeatReceiver=None, mpyClient=None):
    mpycb = mpyStreamCallback
    if type(context) is str:
        domain = context
    elif type(context) is InstanceObject or ActorObject:
        domain = context.domain
    elif type(context) is AccountObject:
        domain = context.domain
        mpycb = mpyUserStreamCallback
    else:
        print("WARN! startStreamingThread given context of type %s" % type(context))

    if mpyClient is None:
        try:
            mpyClient = await context.mkmpyClient()
        except AttributeError:
            print("WARN! Streaming %s without that object's mpy/httpx client!" % str(context))
            mpyClient = await Mastodon(api_base_url=domain)

    streamingObject = await getattr(mpyClient, 'stream_%s' % streamType)(mpycb(context,
                                                                               statusReceiver,
                                                                               heartbeatReceiver=heartbeatReceiver),
                                                                         run_async=True)
    fp_globals.activeStreamingObjects.add(streamingObject)
    return streamingObject


async def mastodonTimeline(timeline='home', **kw):
    from fe_core.fe_types import ActorObject

    context = kw["context"]
    if type(context) == str:
        try:
            api_base_url = context.split("@")[-1]
        except IndexError:
            api_base_url = context
    else:
        api_base_url = context.domain

    try:
        mpyClient = await context.mkmpyClient()
    except AttributeError:
        try:
            account = kw["account"]
            mpyClient = account.mpyInstance
        except KeyError:
            mpyClient = await Mastodon(api_base_url=api_base_url)

    if timeline in ['home', 'local', 'public', 'user']:
        mpyTimelineList = await mpyClient.timeline(timeline=timeline)

    elif type(context) == ActorObject:
        try:
            actorId = context.ids[api_base_url]
        except KeyError:
            print("Actor in memory, but their id for instance %s isn't." % api_base_url)
            #TODO: This code should be moved to the actor object
            # Not in DB. Try to get it from actor's home server
            #May only work on pleroma instances
            actorSearchResults = await mpyClient.account_search(context.uri)
            for i in actorSearchResults:  # mpy user dicts
                if i.url == context.uri:
                    ##Actor wasn't previously in DB. Adding.
                    aObject = mpyUserDictToActorObject(i, api_base_url)
                    actorId = aObject.ids[api_base_url]

        mpyTimelineList = await mpyClient.account_statuses(actorId, **kw)

    nativeList = []
    for i in reversed(mpyTimelineList):
        nativeList.append(mpyTootDictToStatusObject(i, api_base_url))
    else:
        try:
            kw["callback"](nativeList, mpyClient)
        except KeyError:
            return nativeList, mpyClient
        except Exception as e:
            print("!!!!Failure trying to callback! %s" % repr(e))