#!/usr/bin/python
from asyncio import create_task, get_event_loop
from datetime import datetime
from PyQt5.QtCore import (Qt)
from PyQt5.QtWidgets import QPushButton, QVBoxLayout, \
    QHBoxLayout, QLabel, QPlainTextEdit, QDialog, QDialogButtonBox, QButtonGroup, \
    QRadioButton, QLineEdit, QCheckBox, QStackedWidget, QTabWidget, QWidget, QGroupBox
from PyQt5.QtGui import QTextCursor

from . import widgets
from .widgets import accountSelector, accountSelectorComboBox
from .tabs import mastodon_tabs
from fe_core import fp_globals, fp_sqlite, discoverInstance
from fe_core.fe_types.account import AccountObject
from fe_core.workers import kickoffAccountWorker


class actionFromAccountsDialog(QDialog):

    def __init__(self, parent, actionText=None):
        super().__init__(parent)
        if not actionText:
            self.actionText = "Do Thing"
        else:
            self.actionText = actionText
        self.setWindowTitle(self.actionText)
        self.buttonBox = QDialogButtonBox(Qt.Horizontal)

        self.cancelButton = QPushButton("Cancel")
        self.postButton = QPushButton(self.actionText)
        self.postButton.clicked.connect(self.doThing)
        self.buttonBox.addButton(self.postButton, QDialogButtonBox.ButtonRole.AcceptRole)
        self.buttonBox.addButton(self.cancelButton, QDialogButtonBox.ButtonRole.RejectRole)

        self.layout = QVBoxLayout()
        self.initUI(self.layout)

        self.accountSelector = accountSelector(self)
        self.accountSelector.setMaximumHeight(100)

        self.layout.addWidget(QLabel("%s from these accounts:" % self.actionText))
        self.layout.addWidget(self.accountSelector)
        self.layout.addWidget(self.buttonBox)
        self.setLayout(self.layout)
        self.show()

    def initUI(self, layout):
        # This is to be overridden by subclasses
        # It is called after the dialog is set up
        # After it returns, the account list is added
        pass

    def doThing(self):
        # This is to be overridden by subclasses
        # It is called when the user presses the button
        pass

    def getSelectedAccountObjects(self):
        selectedAccounts = []
        for key, value in self.accountSelector.accountItems.items():
            if value.checkState():
                selectedAccounts.append(fp_globals.loadedAccounts[key])
        return selectedAccounts

class followFromAccountsDialog(actionFromAccountsDialog):
    def __init__(self, parent, actor):
        super().__init__(parent, "Follow %s" % actor.handle)
        self.actor = actor

    def doThing(self):
        for i in self.getSelectedAccountObjects():
            get_event_loop().create_task(i.followActor(self.actor))
        self.accept()


class likeFromAccountsDialog(actionFromAccountsDialog):
    def __init__(self, parent, status):
        super().__init__(parent, "Like Status")
        self.status = status

    def doThing(self):
        for i in self.getSelectedAccountObjects():
            get_event_loop().create_task(i.likeStatus(self.status))
        self.accept()


class boostFromAccountsDialog(actionFromAccountsDialog):
    def __init__(self, parent, status):
        super().__init__(parent, "Boost Status")
        self.status = status

    def doThing(self):
        for i in self.getSelectedAccountObjects():
            get_event_loop().create_task(i.boostStatus(self.status))
        self.accept()


class shuttingDownDialog(QDialog):
    def __init__(self, parent, *args, **kwargs):
        super().__init__(parent, *args, **kwargs)
        self.setWindowTitle("Shutting down cleanly...")
        self.initUI()

    def initUI(self):
        self.resize(300, 100)
        self.layout = QVBoxLayout()
        self.layout.addWidget(QLabel("Saving state and shutting down."))
        self.layout.addWidget(QLabel("This may take up to 10 seconds..."))
        self.setLayout(self.layout)
        self.show()

class newVersionAvailableDialog(QDialog):
    def __init__(self, parent, latestVersion, *args, **kwargs):
        super().__init__(parent, *args, **kwargs)
        self.lv = latestVersion
        self.setMinimumSize(600,200)
        self.setWindowTitle("This copy of %s blows and you should get a new one." % fp_globals.productName)
        self.initUI()

    def initUI(self):
        self.buttonBox = QDialogButtonBox(QDialogButtonBox.Ok)
        self.popBrowserButton = QPushButton("Click here to download it in your default browser.")
        self.buttonBox.accepted.connect(self.accept)
        self.popBrowserButton.clicked.connect(self.openBrowserTabAction)
        self.layout = QVBoxLayout()
        self.layout.addWidget(QLabel("A new version of %s is available." % fp_globals.productName))
        self.layout.addWidget(QLabel("Version %i was posted on %s" % (self.lv["version"], datetime.fromtimestamp(self.lv["timestamp"]).strftime("%A, %d %b %Y %l:%M %p"))))
        self.layout.addWidget(self.popBrowserButton)
        self.layout.addStretch()
        self.layout.addWidget(QLabel("If that doesn't work, annoy r000t about it."))
        self.layout.addWidget(self.buttonBox)
        self.setLayout(self.layout)
        self.show()

    def openBrowserTabAction(self):
        from webbrowser import open as openBrowserTab
        openBrowserTab("https://fediengine.org/download/fediengine_%s%i.tar.gz" % (
        fp_globals.productChannel, self.lv["version"]))
        self.accept()


class globalSettingsWindow(QDialog):
    def __init__(self, parent, *args, **kwargs):
        super().__init__(parent, *args, **kwargs)
        self.setWindowTitle("Configure %s" % fp_globals.productName)
        self.initUI()

    def initUI(self):
        self.resize(700, 500)
        self.buttonBox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel | QDialogButtonBox.Apply)
        self.buttonBox.accepted.connect(self.saveAllPages)
        self.buttonBox.rejected.connect(self.reject)

        self.settingsPaneHolderLayout = QVBoxLayout()
        self.settingsPaneHolderWidget = QStackedWidget()
        self.settingsPaneHolderLayout.addWidget(self.settingsPaneHolderWidget)
        self.settingsPaneHolderLayout.addWidget(self.buttonBox)
        self.globalSettingsTreeWidgdet = fp_widgets.globalSettingsTreeWidget(self.settingsPaneHolderWidget)
        self.globalSettingsTreeWidgdet.currentItemChanged.connect(self.selectedSettingsTreeWidgetItemChanged)

        self.layout = QHBoxLayout()
        self.layout.addWidget(self.globalSettingsTreeWidgdet)
        self.layout.addLayout(self.settingsPaneHolderLayout)
        self.setLayout(self.layout)
        self.show()

    def saveAllPages(self):
        from fe_core.fp_sqlite import SAVE
        pageCount = self.settingsPaneHolderWidget.count()
        for i in [self.settingsPaneHolderWidget.widget(f) for f in range(pageCount)]:
            i.applyChanges()
        SAVE()
        self.accept()


    def selectedSettingsTreeWidgetItemChanged(self, i):
        self.settingsPaneHolderWidget.setCurrentIndex(self.globalSettingsTreeWidgdet.indexOfTopLevelItem(i))


class newStatusDialog(QDialog):
    def __init__(self, parent, *args, **kw):
        super().__init__(parent)
        self.setWindowTitle("Post New Status")
        self.initUI(**kw)

    def initUI(self, inReplyTo=None, quote=False, accounts=None):
        self.inReplyTo = inReplyTo
        self.quote = quote
        self.buttonBox = QDialogButtonBox(Qt.Horizontal)

        self.cancelButton = QPushButton("Cancel")
        self.postButton = QPushButton("Post Status")
        self.postButton.clicked.connect(self.postStatus)
        self.buttonBox.addButton(self.postButton, QDialogButtonBox.ButtonRole.AcceptRole)
        self.buttonBox.addButton(self.cancelButton, QDialogButtonBox.ButtonRole.RejectRole)

        self.layout = QVBoxLayout()

        self.contentEditor = QPlainTextEdit()
        if inReplyTo:
            self.layout.addWidget(QLabel("In reply to %s..." % str(inReplyTo.actor)))
            if not self.quote:
                self.contentEditor.setPlainText(str(inReplyTo.actor) + "\n")
                self.contentEditor.moveCursor(QTextCursor.End)
            else:
                self.contentEditor.setPlainText("\n\n%s %s:\n%s" % (str(inReplyTo.actor), inReplyTo.uri, inReplyTo.sanitizedContent))

        self.characterCount = QLabel("0")
        self.contentEditor.textChanged.connect(self.updateCharacterCount)
        self.accountSelector = accountSelector(self, accounts)
        self.accountSelector.setMaximumHeight(100)

        #TODO: Set default combobox option per account visibility scope setitngs
        self.layout.addWidget(fp_widgets.visibilityScopeSelectorComboBox(self))
        self.layout.addWidget(self.contentEditor)
        self.layout.addWidget(self.characterCount)
        self.layout.addWidget(QLabel("Post from these accounts:"))
        self.layout.addWidget(self.accountSelector)
        self.layout.addWidget(self.buttonBox)
        self.setLayout(self.layout)
        self.show()

    def postStatus(self):
        for key, value in self.accountSelector.accountItems.items():
            if value.checkState():
                account = fp_globals.loadedAccounts[key]

                if account.accountType.startswith("mastodon"):
                    mpykw = {}
                    if self.inReplyTo:
                        if not self.quote:
                            mpykw["in_reply_to_id"] = self.inReplyTo
                    get_event_loop().create_task(account.postStatus(self.contentEditor.toPlainText(), **mpykw))
        self.accept()

    def updateCharacterCount(self):
        self.characterCount.setText(str(len(self.contentEditor.toPlainText())))


class addSearchTabDialog(QDialog):
    def __init__(self, parent, *args, **kwargs):
        super().__init__(parent, *args, **kwargs)

        self.setWindowTitle("Search")
        self.initUI()

    def initUI(self):
        QBtn = QDialogButtonBox.Ok
        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.addTab)

        self.layout = QVBoxLayout()
        self.l1layout = QVBoxLayout()
        # TODO: this needs worded better
        self.searchTextBox = QLineEdit()
        self.l1layout.addWidget(QLabel("Search in status content "))
        self.l1layout.addWidget(self.searchTextBox)
        self.layout.addLayout(self.l1layout)
        self.layout.addWidget(self.buttonBox)
        self.setLayout(self.layout)
        self.show()

    def addTab(self):
        from tabs import searchResultTab
        self.parent().tabs.addTab(searchResultTab(self.parent(), self.searchTextBox.text()), "(local) Search: ")
        self.accept()


class addAccountDialog(QDialog):
    def __init__(self, parent, *args, **kwargs):
        super().__init__(parent, *args, **kwargs)
        self.setWindowTitle("Add New Account")
        self.initUI()

    def initUI(self, instance=None):
        self.buttonBox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        self.buttonBox.accepted.connect(self.addAccount)
        self.buttonBox.rejected.connect(self.close)

        self.layout = QVBoxLayout()
        self.l1layout = QVBoxLayout()
        # TODO: this needs worded better
        self.l1layout.addWidget(QLabel("URL to Mastodon or MastodonFE"))

        self.l2layout = QHBoxLayout()
        self.instanceURLTextBox = QLineEdit()
        if instance:
            self.instanceURLTextBox.setText(instance.domain)

        self.l2layout.addWidget(QLabel("https://"))
        self.l2layout.addWidget(self.instanceURLTextBox)
        self.l2layout.addWidget(QLabel("/"))

        self.oauthCodeTextBox = QLineEdit()
        self.oauthCodeTextBox.setPlaceholderText("OAuth Code")
        self.oauthCodeTextBox.setVisible(False)

        self.layout.addLayout(self.l1layout)
        self.layout.addLayout(self.l2layout)
        self.layout.addWidget(self.oauthCodeTextBox)
        self.layout.addWidget(self.buttonBox)
        self.setLayout(self.layout)
        self.show()

    async def backgroundAddAccount(self):
        # TODO: capture errors, prompt user to retry, suggest alternate FE paths
        # TODO: get domain from mastodon.py, not the text box
        self.accessToken = await self.mpyInstance.log_in(code=self.oauthCode)
        localUser = await self.mpyInstance.me()
        handle = '%s@%s' % (localUser.acct, self.instanceURLTextBox.text())
        auth = (self.mpyInstance.client_id, self.mpyInstance.client_secret, self.oauthCode, self.accessToken)
        fp_globals.loadedAccounts[handle] = AccountObject(handle, auth)
        account = fp_globals.loadedAccounts[handle]
        fp_sqlite.addAccount(handle, *auth)
        kickoffAccountWorker(account)
        await account.log_in()
        h = self.parent().tabs.addTab(mastodon_tabs.mastodonHomeTimelineTab(self.parent(), account), "...")
        self.parent().tabs.widget(h).updateTabName()

    def actuallyAddAccount(self):
        self.oauthCode = self.oauthCodeTextBox.text()
        get_event_loop().create_task(self.backgroundAddAccount())
        self.accept()
        self.close()

    async def backgroundMakempyAndPopBrowser(self):
        # TODO: put mastodon specific code elsewhere
        # probably when other types of servers are supported
        from webbrowser import open as openBrowserTab
        from fe_core.external.mastodon import Mastodon

        # per-account ids and secrets are used to simplify the use of multiple accounts
        # it also prevents correlation, enhancing user privacy
        # TODO: optional stealth
        self.mpyInstance = await Mastodon(api_base_url=self.instanceURLTextBox.text())
        client_id, client_secret = self.mpyInstance.create_app("fediEngine",
                                                               api_base_url=self.instanceURLTextBox.text())
        self.mpyInstance.client_secret = client_secret
        self.mpyInstance.client_id = client_id

        openBrowserTab(self.mpyInstance.auth_request_url())

    def addAccount(self):
        get_event_loop().create_task(self.backgroundMakempyAndPopBrowser())
        self.buttonBox.accepted.disconnect()
        self.buttonBox.accepted.connect(self.actuallyAddAccount)
        self.oauthCodeTextBox.setVisible(True)


class marryAmericanWomanDialog(QDialog):
    # I couldn't get this idea for an easter egg out of my head until I wrote it
    # Luckily it only took like 20 minutes
    def __init__(self, parent):
        super().__init__(parent)
        self.parent = parent
        self.buttonBox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        self.buttonBox.accepted.connect(self.marryAmericanWoman)
        self.buttonBox.rejected.connect(self.reject)

        self.setWindowTitle("Marrying an American Woman")

        self.layout = QVBoxLayout()
        self.layout.addWidget(QLabel("Are you ABSOLUTELY sure you want to marry an American woman?"))
        self.layout.addWidget(QLabel("Please consider the possible consequences!"))
        self.layout.addWidget(self.buttonBox)

        self.setLayout(self.layout)
        self.show()

    def marryAmericanWoman(self):
        h = marriedAmericanWomanDialog(self.parent)
        self.accept()


class marriedAmericanWomanDialog(QDialog):
    def __init__(self, parent):
        super().__init__(parent)
        self.buttonBox = QDialogButtonBox(QDialogButtonBox.Ok)
        self.buttonBox.accepted.connect(self.accept)
        self.setWindowTitle("Married an American Woman")
        self.layout = QVBoxLayout()
        self.layout.addWidget(QLabel("You marry an American woman. Some time later, your house is stolen in a divorce."))
        self.layout.addWidget(QLabel("You distinctly remember being warned of this sort of outcome back in 2019."))
        self.layout.addWidget(QLabel("gg"))
        self.layout.addWidget(self.buttonBox)
        self.setLayout(self.layout)
        self.show()

class discoverInstanceDialog(QDialog):
    def __init__(self, parent, *args, **kwargs):
        super().__init__(parent, *args, **kwargs)
        self.setWindowTitle("Discover Instance")
        self.initUI()

    def initUI(self):
        self.textBox = QLineEdit()
        self.textBox.setPlaceholderText("example.org")
        discoverButton = QPushButton("Discover")
        discoverButton.clicked.connect(self.discoverInstanceButtonPressed)
        formLayout = QHBoxLayout()
        formLayout.addWidget(self.textBox)
        formLayout.addWidget(discoverButton)

        self.instanceBasicStatusLabel = QLabel()
        self.addAccountButton = QPushButton("Add Account...")
        self.addAccountButton.clicked.connect(fp_globals.Application.addAccountAction.triggered)
        self.openTimelineTabButton = QPushButton("View Local Timeline")
        [x.setEnabled(False) for x in (self.addAccountButton, self.openTimelineTabButton)]
        buttonLayout = QHBoxLayout()
        buttonLayout.addWidget(self.addAccountButton)
        buttonLayout.addWidget(self.openTimelineTabButton)
        resultsLayout = QVBoxLayout()
        resultsLayout.addWidget(self.instanceBasicStatusLabel)
        resultsLayout.addLayout(buttonLayout)
        self.resultsGroupBox = QGroupBox("Results")
        self.resultsGroupBox.setLayout(resultsLayout)

        self.layout = QVBoxLayout()
        self.layout.addLayout(formLayout)
        self.layout.addWidget(self.resultsGroupBox)
        self.setLayout(self.layout)
        self.show()

    def discoverInstanceButtonPressed(self):
        #TODO: Sanitize input
        entered = self.textBox.text()

        [x.setEnabled(False) for x in (self.addAccountButton, self.openTimelineTabButton)]
        self.instanceBasicStatusLabel.setText("Trying %s..." % entered)
        try:
            instance = create_task(discoverInstance(entered))
            self.instanceBasicStatusLabel.setText("Instance is UP.")
            [x.setEnabled(True) for x in (self.addAccountButton, self.openTimelineTabButton)]
        except Exception as e:
            self.instanceBasicStatusLabel.setText("Instance is DOWN - (Generic error %s)" % repr(e))


class addTabDialog(QDialog):
    def __init__(self, parent, *args, **kwargs):
        super().__init__(parent, *args, **kwargs)
        self.setWindowTitle("New Tab")
        self.initUI()

    def initUI(self):
        self.buttonBox = QDialogButtonBox(QDialogButtonBox.Ok)
        self.buttonBox.accepted.connect(self.addTab)
        self.layout = QVBoxLayout()
        self.l1layout = QVBoxLayout()
        self.l1layout.addWidget(QLabel("This dialog box will be replaced soon."))
        self.tabTypebuttonGroup = QButtonGroup()
        self.tabTypebuttonGroup.setExclusive(True)
        self.mastodonLocalTimeline = QRadioButton("Mastodon Local Timeline")
        self.mastodonFederatedTimeline = QRadioButton("Mastodon Federated Timeline")
        self.accountHomeTimeline = QRadioButton("Account Home Timeline")
        self.tabTypebuttonGroup.addButton(self.mastodonLocalTimeline)
        self.tabTypebuttonGroup.addButton(self.mastodonFederatedTimeline)
        self.tabTypebuttonGroup.addButton(self.accountHomeTimeline)
        self.l1layout.addWidget(self.mastodonLocalTimeline)
        self.l1layout.addWidget(self.mastodonFederatedTimeline)
        self.l1layout.addWidget(self.accountHomeTimeline)

        self.l2layout = QHBoxLayout()
        self.instanceURLTextBox = QLineEdit()
        self.l2layout.addWidget(QLabel("https://"))
        self.l2layout.addWidget(self.instanceURLTextBox)
        self.l2layout.addWidget(QLabel("/"))

        self.l3layout = QHBoxLayout()
        self.accountSelector = accountSelectorComboBox(self)
        self.l3layout.addWidget(self.accountSelector)

        self.l4layout = QHBoxLayout()

        self.streamingAPI = QCheckBox("Use Mastodon streaming API")
        self.streamingAPI.setChecked(True)

        self.tabIsPersistentCheckbox = QCheckBox("Persist tab across restarts")
        self.tabIsPersistentCheckbox.setChecked(True)

        self.layout.addLayout(self.l1layout)
        self.layout.addLayout(self.l2layout)
        self.layout.addWidget(self.accountSelector)
        self.layout.addSpacing(10)
        self.layout.addLayout(self.l3layout)
        self.layout.addLayout(self.l4layout)
        self.layout.addWidget(self.streamingAPI)
        self.layout.addSpacing(15)
        self.layout.addWidget(self.tabIsPersistentCheckbox)
        self.layout.addWidget(self.buttonBox)
        self.setLayout(self.layout)
        self.show()

    def addTab(self):
        tabConfig = {"tabIsPersistent": self.tabIsPersistentCheckbox.isChecked(),
                    "disableStreaming": not self.streamingAPI} #There's NO WAY that works!

        tabClass = {self.mastodonLocalTimeline: mastodon_tabs.mastodonLocalTimelineTab,
                    self.mastodonFederatedTimeline: mastodon_tabs.mastodonFederatedTimelineTab}

        if self.accountSelector.currentText() != "(Anonymous)":
            tabClass[self.accountHomeTimeline] = mastodon_tabs.mastodonHomeTimelineTab
            #tabArgs["account"] = fp_globals.loadedAccounts[self.accountSelector.currentText()]
            targetContext = fp_globals.loadedAccounts[self.accountSelector.currentText()]
        else:
            #tabArgs["domain"] = self.instanceURLTextBox.text()
            #TODO: Try to match a loaded instance, get instance data, or error out before closing window
            targetContext = self.instanceURLTextBox.text()

        h = self.parent().tabs.addTab(tabClass[self.tabTypebuttonGroup.checkedButton()](self, targetContext, tabConfig), "...")
        self.parent().tabs.widget(h).updateTabName()
        self.accept()


class objectDetailsDialog(QDialog):
    title = "Viewer"
    def __init__(self, parent, context, inline=False, *args, **kwargs):
        super().__init__(parent, *args, **kwargs)
        self.layout = QVBoxLayout()
        self.initUI(context, tasksOnBottom=inline)
        self.viewerWidget = QWidget()
        self.viewerWidget.setLayout(self.viewerLayout)
        self.setLayout(self.layout)
        if inline:
            self.layout.addWidget(self.viewerWidget)
        else:
            self.setWindowTitle(self.title)
            self.buttonBox = QDialogButtonBox(QDialogButtonBox.Ok)
            self.buttonBox.accepted.connect(self.accept)
            self.tabWidget = QTabWidget()
            self.tabWidget.addTab(self.viewerWidget, str(type(context)))
            self.settingsPaneWidget = fp_widgets.settingsPaneWidget("Settings", "Change settings specific to %s" % str(self.settingsContext), self.settingsContext.config.values(), True)
            self.tabWidget.addTab(self.settingsPaneWidget, "Settings")
            self.layout.addWidget(self.tabWidget)
            self.layout.addWidget(self.buttonBox)
            self.show()

    def openBoostStatusDialog(self):
        h = boostFromAccountsDialog(self, self.status)

    def openLikeStatusDialog(self):
        h = likeFromAccountsDialog(self, self.status)

    def openReplyToStatusDialog(self):
        h = newStatusDialog(self, inReplyTo=self.status, quote=False)

    def openQuoteReplyStatusDialog(self):
        h = newStatusDialog(self, inReplyTo=self.status, quote=True)

    def openActorDetailsDialog(self):
        h = actorDetailsDialog(self, self.actor)

    def openFollowActorDialog(self):
        h = followFromAccountsDialog(self, self.actor)

    def openActorTab(self):
        from tabs.mastodon_tabs import mastodonActorTab
        tabArgs = {"context": self.actor,
                   "domain": self.actor.domain}
        h = fp_globals.Application.tabs.addTab(mastodonActorTab(fp_globals.Application, **tabArgs), "...")
        fp_globals.Application.tabs.widget(h).updateTabName()

class actorDetailsDialog(objectDetailsDialog):
    title = "Actor Viewer"
    def initUI(self, actor, tasksOnBottom=False):
        self.actor = actor
        if tasksOnBottom:
            self.actionLayout = QHBoxLayout()
        else:
            self.actionLayout = QVBoxLayout()

        self.viewerLayout = QHBoxLayout()
        self.detailsLayout = QVBoxLayout()
        self.detailsWidget = fp_widgets.actorDetailsWidget(self, self.actor)
        self.detailsLayout.addWidget(self.detailsWidget)
        self.viewerLayout.addLayout(self.detailsLayout)
        if tasksOnBottom:
            self.detailsLayout.addLayout(self.actionLayout)
        else:
            self.viewerLayout.addLayout(self.actionLayout)

        self.actorTasksGroupBox = fp_widgets.actorTasksGroupBox(self)
        self.actionLayout.addWidget(self.actorTasksGroupBox)
        self.actionLayout.addStretch()
        self.settingsContext = self.actor.instance


class statusDetailsDialog(objectDetailsDialog):
    title = "Status Viewer"
    def initUI(self, status, tasksOnBottom=False):
        self.status = status
        self.actor = status.actor
        if tasksOnBottom:
            self.actionLayout = QHBoxLayout()
        else:
            self.actionLayout = QVBoxLayout()

        self.viewerLayout = QHBoxLayout()
        self.detailsLayout = QVBoxLayout()
        self.detailsWidget = widgets.statusDetailsWidget(self, self.status)
        self.detailsLayout.addWidget(self.detailsWidget)

        self.statusContextView = widgets.statusListDataView()
        self.statusContextView.addStatus(status)
        self.statusContextView.updateModel()
        self.detailsLayout.addWidget(self.statusContextView)

        self.viewerLayout.addLayout(self.detailsLayout)
        if tasksOnBottom:
            self.detailsLayout.addLayout(self.actionLayout)
        else:
            self.viewerLayout.addLayout(self.actionLayout)

        self.statusTasksGroupBox = widgets.statusTasksGroupBox(self)
        self.actionLayout.addWidget(self.statusTasksGroupBox)

        self.actorTasksGroupBox = widgets.actorTasksGroupBox(self)
        self.actionLayout.addWidget(self.actorTasksGroupBox)
        if not tasksOnBottom:
            self.actionLayout.addStretch()
        self.settingsContext = self.status.actor.instance
